#!/usr/bin/python3
import sys
import os
import configparser
import sqlite3
import subprocess
import getpass
import json
from json import JSONDecodeError
import re
from datetime import datetime,timedelta

def refresh_uu_config(new_config_file,old_config_file):
    config_new=configparser.ConfigParser(allow_no_value=True)
    config_new.optionxform = str
    config_old=configparser.ConfigParser(allow_no_value=True)
    config_old.optionxform = str
    config_new.read(new_config_file)
    config_old.read(old_config_file)
    for section in config_new.sections():
        if not config_old.has_section(section):
            config_old.add_section(section)
            print("new section %s added"%(section))  
        for option in config_new.options(section):
            if not config_old.has_option(section,option):
                value = config_new[section][option]
                config_old.set(section,option,value)
                print("new option:%s %s"%(option,value))
    with open(old_config_file,'w') as f:
        config_old.write(f)

def setconfvalue(section,key,value):
    cmd="kconf2 set %s %s %s"%(str(section),str(key),str(value))
    print(cmd)
    ret=subprocess.run(cmd,shell=True)
    return ret.returncode
    
db_file="/var/lib/kylin-system-updater/kylin-system-updater.db"
if (os.path.exists(db_file)):
    conn = sqlite3.connect(db_file)
    cursor = conn.cursor()
    cursor.execute('SELECT auto_check FROM display')
    dat = cursor.fetchone()
    auto_check=dat[0]
    print("auto_check:%s"%auto_check)

def refresh_frontend_yaml():
    config_file = "/var/lib/unattended-upgrades/unattended-upgrades-policy.conf"
    db_file="/var/lib/kylin-system-updater/kylin-system-updater.db"
    experience_config_file="/var/lib/kylin-software-properties/config/updateExperience.conf"
    server_address_config_file="/var/lib/kylin-software-properties/template/kylin-source-update.conf"
    if (os.path.exists(db_file)):
        conn = sqlite3.connect(db_file)
        cursor = conn.cursor()
        cursor.execute('SELECT auto_check FROM display')
        dat = cursor.fetchone()
        auto_check=dat[0]
        print("auto_check:%s"%auto_check)
        setconfvalue("kylin-update-frontend.notification","enable",auto_check)
        cursor.execute('SELECT download_limit FROM display')
        dat = cursor.fetchone()
        download_limit=dat[0]
        print("download_limit:%s"%download_limit)
        setconfvalue("kylin-update-frontend.downloadlimit","enable",download_limit)
        cursor.execute('SELECT download_limit_value FROM display')
        dat = cursor.fetchone()
        download_limit_value=dat[0]
        print("download_limit_value:%s"%download_limit_value)
        setconfvalue("kylin-update-frontend.downloadlimit","value",download_limit_value)
        cursor.close()
        conn.close()     
    if (os.path.exists(config_file)):
        config=configparser.ConfigParser(allow_no_value=True)
        config.optionxform = str
        config.read(config_file)
        dat=config['autoUpgradePolicy']['autoUpgradeState']
        auto_download="true"
        if(dat=="on"):
            auto_download="true" 
        elif(dat=="off"):
            auto_download="false"
        print("auto_download:%s"%auto_download)
        setconfvalue("kylin-update-frontend.autodownload","enable",auto_download)              
        dat=config['autoUpgradePolicy']['updateDays']
        print("update days:%s"%dat)
        setconfvalue("kylin-update-frontend.downloadsettings","period",dat)
        worktime="true"       
        user=os.getlogin()
        if os.path.exists("/home/%s/.config/durationcheckbox"%user):
            worktime="true"
        else:
            worktime="false"
        # dat=config['autoUpgradePolicy']['workTime']
        # worktime="true"
        # if(dat=="on"):
        #     worktime="true" 
        # elif(dat=="off"):
        #     worktime="false"
        print("work time:%s"%worktime)
        setconfvalue("kylin-update-frontend.downloadsettings","designated",worktime)
        dat=config['autoUpgradePolicy']['downloadTime']
        print("download time:%s"%dat)
        setconfvalue("kylin-update-frontend.downloadsettings","designatedtime",dat)
    if (os.path.exists(experience_config_file)):
        config=configparser.ConfigParser(allow_no_value=True)
        config.optionxform = str          
        config.read(experience_config_file)
        dat=config['Experience']['isUpdateExperience']
        print("betaversion:%s"%dat)
        setconfvalue("kylin-update-frontend.betaversion","enable",dat) 
    if (os.path.exists(server_address_config_file)):
        config=configparser.ConfigParser(allow_no_value=True)
        config.optionxform = str
        config.read(server_address_config_file)
        dat=config['Server']['Address']
        print("server address:%s"%dat)
        setconfvalue("kylin-update-frontend.serversettings","address",dat) 
        dat=config['Server']['Port']
        print("server port:%s"%dat)
        setconfvalue("kylin-update-frontend.serversettings","port",dat) 

def safe_merge_json(file1, file2):
    """安全合并两个JSON文件"""
    def load_file(path):
        try:
            with open(path, "r", encoding="utf-8") as f:
                return json.load(f)
        except FileNotFoundError:
            raise RuntimeError(f"文件 {path} 不存在")
        except JSONDecodeError as e:
            raise ValueError(f"文件 {path} 格式错误：{e.msg} (行{e.lineno}列{e.colno})")
    
    try:
        data1 = load_file(file1)
        data2 = load_file(file2)
        
        # 深度合并逻辑
        def merge(dict1, dict2):
            for k, v in dict2.items():
                if isinstance(v, dict) and k in dict1:
                    merge(dict1[k], v)
                else:
                    dict1[k] = v
            return dict1
        
        merged = merge(data1, data2)
        
        with open(file2, "w", encoding="utf-8") as f:
            json.dump(merged, f, indent=2)
            
    except Exception as e:
        print(f"合并失败：{str(e)}")
        raise

def ini_to_json(ini_path: str) -> dict:
    """
    将INI配置文件转换为JSON格式的字典对象
    
    Args:
        ini_path: INI文件路径
        
    Returns:
        dict: 转换后的JSON格式字典
        
    Raises:
        FileNotFoundError: 当INI文件不存在时抛出
        ValueError: 当文件格式错误时抛出
    """
    # 初始化配置解析器（网页1、网页4、网页5的核心方法）
    config = configparser.ConfigParser()
    
    # 读取INI文件（网页5的基础操作）
    if not config.read(ini_path):
        raise FileNotFoundError(f"文件 {ini_path} 未找到")
    
    # 构建嵌套字典（网页1、网页4的转换逻辑）
    result = {}
    for section in config.sections():
        result[section] = dict(config.items(section))  # 转换section为字典结构
        
    return result

def parse_time_range(time_str: str) -> dict:
    """
    解析时间字符串并生成时间范围字典
    
    参数：
    time_str - 输入时间字符串，支持三种格式：
               "HH:MM"、"HH:MM:SS"、"HH:MM-HH:MM"
    
    返回：
    {'start': 'HH:MM', 'end': 'HH:MM'}
    
    异常：
    ValueError - 当输入格式无效或时间值越界时抛出
    """
    # 格式验证正则表达式
    time_pattern = r'^([0-2][0-9]:[0-5][0-9])(?::[0-5][0-9])?(?:-([0-2][0-9]:[0-5][0-9]))?$'
    
    if match := re.match(time_pattern, time_str):
        start_time, end_time = match.groups()
        
        # 处理第三种格式（时间区间）
        if end_time:
            return validate_and_format(start_time, end_time)
        
        # 处理前两种格式（单个时间）
        base_time = time_str[:5]  # 截取HH:MM部分
        calculated_end = calculate_end_time(base_time)
        return {'start': base_time, 'end': calculated_end}
    
    raise ValueError(f"无效时间格式：{time_str}")

def validate_and_format(start: str, end: str) -> dict:
    """验证并格式化时间区间"""
    try:
        # 时间有效性验证
        datetime.strptime(start, '%H:%M')
        datetime.strptime(end, '%H:%M')
        return {'start': start, 'end': end}
    except ValueError as e:
        raise ValueError(f"无效时间值：{str(e)}")

def calculate_end_time(base: str) -> str:
    """计算3小时后时间（处理跨天）"""
    try:
        # 时间解析与计算
        base_dt = datetime.strptime(base, '%H:%M')
        end_dt = base_dt + timedelta(hours=3)
        return end_dt.strftime('%H:%M')
    except ValueError as e:
        raise ValueError(f"无效基准时间：{str(e)}")

def safe_merge_json_to_file(d: dict, file_path: str) -> dict:
    """
    将字典安全合并到JSON文件，覆盖同名键值
    
    参数：
    d - 需要合并的字典
    file_path - 目标JSON文件路径
    
    返回值：
    dict - 合并后的完整字典
    
    异常：
    RuntimeError - 包含具体错误描述(文件不存在/权限问题/格式错误等)
    
    功能特性：
    - 支持嵌套字典的深度合并
    - 自动创建空文件(当文件不存在时)
    - 原子化写入防止数据损坏
    - 详细的错误定位信息
    """
    # 异常预处理
    if not isinstance(d, dict):
        raise TypeError("输入参数d必须是字典类型")

    target_data = {}
    
    try:
        # 阶段1：安全读取目标文件
        if os.path.exists(file_path):
            with open(file_path, 'r', encoding='utf-8') as f:
                try:
                    target_data = json.load(f)
                    if not isinstance(target_data, dict):
                        raise ValueError("JSON文件顶层结构必须为字典")
                except json.JSONDecodeError as e:
                    # 增强错误信息可读性
                    err_msg = (
                        f"文件格式错误: {e.msg}\n"
                        f"错误位置: 第{e.lineno}行第{e.colno}列\n"
                        f"错误上下文: {e.doc[e.pos-20:e.pos+20]}"
                    )
                    raise json.JSONDecodeError(err_msg, e.doc, e.pos)
        else:
            os.makedirs(os.path.dirname(file_path), exist_ok=True)
            
        # 阶段2：递归深度合并
        def recursive_merge(base: dict, update: dict) -> None:
            """嵌套字典合并算法"""
            for k, v in update.items():
                if isinstance(v, dict) and k in base and isinstance(base[k], dict):
                    recursive_merge(base[k], v)
                else:
                    base[k] = v
        
        recursive_merge(target_data, d)
        
        # 阶段3：原子化写入
        temp_path = f"{file_path}.tmp"
        with open(temp_path, 'w', encoding='utf-8') as f:
            json.dump(target_data, f, indent=4, ensure_ascii=False)
        
        os.replace(temp_path, file_path)
        return target_data

    except PermissionError as e:
        raise RuntimeError(f"文件权限错误: {e.strerror}") from e
    except IOError as e:
        raise RuntimeError(f"磁盘IO错误: {e.strerror}") from e
    except Exception as e:
        if os.path.exists(temp_path):
            os.remove(temp_path)  # 清理临时文件
        raise RuntimeError(f"合并失败: {str(e)}") from e

def merge_ini_json(ini_file,json_file):
    json_data=ini_to_json(ini_file)
    print(json_data)
    data={}
    autoupgradepolicy=json_data.get('autoUpgradePolicy',{})
    updatedays=int(autoupgradepolicy.get('updatedays',1))
    randomrange=int(autoupgradepolicy.get('randomrange',1))
    if(autoupgradepolicy.get('autoupgradestate','off')=='on'):
        data.update({'updatemode':'autoupdate'})
        downloadmode=autoupgradepolicy.get('downloadmode','timing')
        downloadtime=parse_time_range(autoupgradepolicy.get('downloadtime','11:00-12:00'))
        installmode=autoupgradepolicy.get('installmode','timing')
        installtime=parse_time_range(autoupgradepolicy.get('installtime','10:00-11:00'))
        autoupdate={}
        if(downloadmode=='timing'):
            autoupdate.update({'downloadtime':[downloadtime]})
        if(installmode=='timing'):
            autoupdate.update({'mode':'timing'})
            autoupdate.update({'installtime':installtime.get('start',"10:00")})
        elif(installmode=='bshutdown'):
            autoupdate.update({'mode':'shutdown'})
        autoupdate.update({'downloadtimerandom': randomrange})
        autoupdate.update({'period':updatedays})
        data.update({'autoupdate':autoupdate})
    elif(autoupgradepolicy.get('predownload','off')=='on'):
        data.update({'updatemode':'predownload'})
        predownloadtime=parse_time_range(autoupgradepolicy.get('predownloadtime','10:00-11:00'))
        predownload={}
        predownload.update({"downloadtime":[predownloadtime]})
        predownload.update({'downloadtimerandom': randomrange})
        predownload.update({'period':updatedays})
        data.update({"predownload":predownload})
    else:
        data.update({'updatemode':'off'})
    db_file="/var/lib/kylin-system-updater/kylin-system-updater.db"
    if (os.path.exists(db_file)):
        conn = sqlite3.connect(db_file)
        cursor = conn.cursor()
        cursor.execute('SELECT auto_check FROM display')
        dat = cursor.fetchone()
        auto_check=dat[0]
        print("auto_check:%s"%auto_check)
        if(auto_check=="true"):
            data.update({'notification':True})
        else:
            data.update({'notification':False})
    safe_merge_json_to_file(data,json_file)

if __name__ == "__main__":
    new_config_file = "/usr/share/kylin-system-updater/unattended-upgrades-policy.conf"
    old_config_file = "/var/lib/unattended-upgrades/unattended-upgrades-policy.conf"
    db_file="/var/lib/kylin-system-updater/kylin-system-updater.db"
    print("argv length:%d"%len(sys.argv))
    if len(sys.argv) >=4:     
        new_config_file=sys.argv[1]
        old_config_file=sys.argv[2]
        db_file=sys.argv[3]
    try:
        refresh_uu_config(new_config_file,old_config_file)
        refreshyamlflag="/var/lib/conf2-refresh-flag"
        if(os.path.exists(refreshyamlflag)):
            pass
        else:
            print("refresh conf2 db")
            refresh_frontend_yaml()
            with open(refreshyamlflag,'w') as f:
                f.write("already refreshed")
            merge_ini_json("/var/lib/unattended-upgrades/unattended-upgrades-policy.conf",'/var/lib/unattended-upgrades/kylin-update-strategy.json')
            # safe_merge_json('/var/cache/kylin-system-updater/rate_limites.json','/var/lib/unattended-upgrades/kylin-update-strategy.json')
    except Exception as e:
        print(e)
        os._exit(0)


